/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Packet.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "Packet.hpp"
#include "logging.hpp"
#include <sstream>
#include <iomanip>
#ifndef EPS_WINDOWS
	#include <net/ethernet.h>
#endif


EPS::Packet::Packet(const uint8_t *ptr, const size_t &skip) :
	data(ptr),
	number_of_bytes_to_skip(skip)
{
	return;
}


const EPS::EthernetHeader *EPS::Packet::get_ethernet_header(void) const
{
	return reinterpret_cast<const EPS::EthernetHeader*>(this->data + this->number_of_bytes_to_skip);
}


const EPS::Ipv4Header *EPS::Packet::get_ipv4_header(void) const
{
	if (not this->is_ipv4())
	{
		return nullptr;
	}

	const uint8_t *ptr = this->data + this->number_of_bytes_to_skip + sizeof(EPS::EthernetHeader);

	return reinterpret_cast<const EPS::Ipv4Header*>(ptr);
}


bool EPS::Packet::is_ipv4(void) const
{
	return (ntohs(this->get_ethernet_header()->ethertype) == ETHERTYPE_IP);
}


bool EPS::Packet::is_ipv4_tcp(void) const
{
	const EPS::Ipv4Header *ipv4_header = this->get_ipv4_header();
	if (ipv4_header && ipv4_header->protocol == IPPROTO_TCP)
	{
		return true;
	}

	return false;
}


void EPS::Packet::debug(void) const
{
	const EPS::EthernetHeader *ethernet_header	= this->get_ethernet_header();
	const bool is_ip							= this->is_ipv4();
	const EPS::Ipv4Header *ipv4_header			= this->get_ipv4_header();

	std::stringstream ss;

	ss	<< "packet debug:"													<< std::endl
		<< "ethernet header:"												<< std::endl
		<< "\tethernet header size: " << sizeof(EPS::EthernetHeader)		<< std::endl
		<< "\tethertype:            0x"		<< std::setfill('0')
											<< std::setw(4)
											<< std::hex
											<< ntohs(ethernet_header->ethertype)
											<< std::dec
											<< " (" << EPS::proto_to_string(ntohs(ethernet_header->ethertype)) << ")"
											<< std::endl
		<< "\tis ipv4:              " << is_ip								<< std::endl;

	if (not ipv4_header)
	{
		ss << "\tno ipv4 header available";
	}
	else
	{
		ss	<< "ipv4 header:"														<< std::endl
			<< "\tip header size:       " << sizeof(EPS::Ipv4Header)				<< std::endl
			<< "\tprotocol:             0x"		<< std::setfill('0')
												<< std::setw(2)
												<< std::hex
												<< (int)ipv4_header->protocol
												<< std::dec							<< std::endl
			<< "\tsrc:                  " << EPS::IPv4(ntohl(ipv4_header->src_ip))	<< std::endl
			<< "\tdst:                  " << EPS::IPv4(ntohl(ipv4_header->dst_ip))	<< std::endl
			<< "\tis tcp:               " << this->is_ipv4_tcp();
	}

	LOG_TRACE << ss.str();

	return;
}
