/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Packet.hpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#pragma once

#include "common.hpp"


namespace EPS
{
	typedef uint8_t MAC[6];

	#pragma pack(push, 1)

	/// typical 14-byte ethernet header.  (Ignores 802.1Q.)
	struct EthernetHeader
	{
		MAC			dst;
		MAC			src;
		// skip the optional 802.1Q (vlan tag)
		uint16_t	ethertype;	///< network byte order, e.g., ETHERTYPE_IP, ETHERTYPE_ARP, or ETHERTYPE_IPV6
	};


	/// typical 20-byte IPV4 header.
	struct Ipv4Header
	{
		uint8_t		ver_ihl;
		uint8_t		dscp_ecn;
		uint16_t	total_length;
		uint16_t	id;
		uint16_t	flags_fragment;
		uint8_t		ttl;
		uint8_t		protocol;	///< e.g., IPPROTO_ICMP, IPPROTO_TCP, or IPPROTO_UDP
		uint16_t	checksum;
		uint32_t	src_ip;		///< network byte order
		uint32_t	dst_ip;		///< network byte order
	};

	#pragma pack(pop)


	/// Simple class to dissect a packet and look for a few key identifiers, such as is_ipv4() and is_ipv4_tcp().
	class Packet final
	{
		public:

			/// Constructor.
			Packet(const uint8_t *ptr, const size_t &skip = 0);

			/// Find the ethernet header.
			const EPS::EthernetHeader *get_ethernet_header(void) const;

			/// Find the ipv4 header. @return nullptr if the packet is not ipv4
			const EPS::Ipv4Header *get_ipv4_header(void) const;

			/// Determine if the packet is IPv4.  (ethertype == ETHERTYPE_IP)
			bool is_ipv4(void) const;

			/// Determine if the packet is IPV4+TCP.  (protocol == IPPROTO_TCP)
			bool is_ipv4_tcp(void) const;

			/// Log some information on this packet.
			void debug(void) const;

			// members

			const uint8_t *	data;						///< pointer to start of packet data
			const size_t	number_of_bytes_to_skip;	///< @see IFF_NO_PI
	};
}
