/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: comm.hpp 2182 2017-03-26 08:30:05Z stephane $
 * See license.txt.
 */

#pragma once

#include "common.hpp"
#include <boost/uuid/uuid.hpp>
#include <string>


namespace EPS
{
	/// Several messages contain a limited vector of addresses.  The maximum number of addresses we can send is controlled with this define.
	#define MAX_NUMBER_OF_OTHER_ADDRESSES 3

	/** Protocol identifier.  This is what determines if two different versions of %EPS can communicate.  This value
	 * must be incremented every time a change is made to the communication protocol that would prevent %EPS from
	 * working correctly with prior versions.
	 */
	#define EPS_PROTOCOL_VERSION	4

	/// Communication structure types.
	enum ECommType
	{
		kUnknown	= 0,
		kMinValid	= 1,
		kHello1		= kMinValid,	///< see @ref EPS::CommHello1 (no longer supported, replaced by @ref EPS:CommHello2)
		kAddress1	= 2,			///< see @ref EPS::CommAddress1
		kData1		= 3,			///< see @ref EPS::CommData1
		kList1		= 4,			///< see @ref EPS::CommList1 (no longer supported, replaced by @ref EPS::CommList2)
		kStats1		= 5,			///< see @ref EPS::CommStats1
		kHello2		= 6,			///< see @ref EPS::CommHello2 (introduced in protocol version 4)
		kList2		= 7,			///< see @ref EPS::CommList2 (introduced in protocol version 4)
		/// remember to update @ref EPS::comm_type_to_string() if you add a new comm type
		kMax
	};

	#pragma pack(push,1)

	/** Communication header:  common fields to all communication commands.  It is padded out to 16 bytes so it is
	 * exactly 1 block once we AES-encrypt it.  This makes it easier when it comes time to read packets off the wire
	 * if we know the common header is exactly 1 AES cipher block.
	 */
	struct CommHeader1 final
	{
		/// reserved for future use, and to ensure the common header is exactly 16 bytes in length @{
		uint32_t	reserved_1;
		uint32_t	reserved_2;
		uint16_t	reserved_3;
		uint8_t		reserved_4;
		/// @}

		uint8_t		protocol_version;	///< protocol version identifier
		uint16_t	number_of_bytes;	///< number of bytes in the comm structure including every byte in the header structure (stored in network byte order)
		uint8_t		comm_type;			///< structure type @see EPS::ECommType @see EPS::comm_type_to_string()

		/** Reserved for future use.  Try and leave this last reserved value at zero, so that when we decrypt the
		 * header bytes (the first 16 bytes of the message) we don't mistake this 16th byte as a PKCS #5 / PKCS #7
		 * padding value and accidentally truncate the header.  @see EPS::Cipher::decrypt()
		 */
		uint8_t		reserved_5;

		/// Constructor.
		CommHeader1(void);

		/// Initialize the header.
		void set(const EPS::ECommType type, const size_t bytes);

		/// A safe way to get @ref number_of_bytes, automatically converting it to host order and resizing it as a size_t.
		size_t size_of_packet(void) const;

		/// A safe way to get @ref comm_type as a EPS::ECommType value.
		EPS::ECommType type(void) const;
	};


	/// Communication header:  initial "hello" sent between clients and servers when they first connect.
	struct CommHello2 final
	{
		static const size_t max_number_of_other_addresses;

		CommHeader1	header;					///< common header
		uint8_t		mode;					///< @see EPS::EMode
		uint32_t	address_eps;			///< eps address (or 0.0.0.0 if it hasn't yet been assigned)
		uint32_t	other_addresses[MAX_NUMBER_OF_OTHER_ADDRESSES]; ///< other addresses this host can use (LAN, WAN, ...) in network byte order
		uint32_t	eps_uptime;				///< timestamp when EPS was started (network byte order)
		uint32_t	system_uptime;			///< timestamp when the system was started (network byte order)
		char		version_str		[20];	///< text representation of the version string
		char		uuid_str		[40];	///< uuid
		char		hostname_str	[40];	///< first 40 chars of the hostname (which can be up to 255 bytes long)
		char		cipherhash_str	[20];	///< nonesense hash used to quickly determine if the cipher key and iv are correctly set
		char		os_version_str	[40];	///< version string of the operating system @see EPS::get_os_name()

		/// Constructor. @{
		CommHello2(void);
		CommHello2(const UUID &uuid, const EPS::EMode mode, const time_t eps_timestamp, const std::string &cipherhash, const IPv4Set &other_addresses, const IPv4 &address_eps = IPv4());
		/// @}

		/// Set the CommHello2 fields.
		void set(const UUID &uuid, const EPS::EMode mode, const time_t eps_timestamp, const std::string &cipherhash, const IPv4Set &other_addresses, const IPv4 &address_eps = IPv4());
	};

	/// Communication header:  send a number of important addresses sent by the server to the clients on startup.
	struct CommAddress1 final
	{
		CommHeader1	header;				///< common header
		uint32_t	network_address;	///< %EPS network
		uint32_t	network_netmask;	///< %EPS netmask
		uint32_t	client_address;		///< %EPS client address

		/// Constructor. @{
		CommAddress1(void);
		CommAddress1(const IPv4 &network, const IPv4 &netmask, const IPv4 &address);
		/// @}

		/// Set the CommAddress1 fields.
		void set(const IPv4 &network, const IPv4 &netmask, const IPv4 &address);
	};

	/// Communication header:  data packet.
	struct CommData1 final
	{
		CommHeader1	header;				///< common header
		UUID16		dst_uuid;			///< destination uuid (if known)
		uint8_t		data[0];			///< start of encapsulated ethernet frame

		/// Initialize the header.
		void set(const size_t number_of_data_bytes, const UUID &uuid);
	};

	#define EPS_HOSTNAME_LENGTH 40
	/** Represents a single %EPS client.  An array of @ref CommList1Data is sent to each peer by the %EPS server as
	 * part of @ref CommList1 whenever a new device joins the %EPS network.
	 */
	struct CommList2Data final
	{
		UUID16		uuid;				///< uuid for the client
		uint32_t	address_eps;		///< eps0 address (network order)
		uint32_t	other_addresses[MAX_NUMBER_OF_OTHER_ADDRESSES]; ///< other addresses this host can use (LAN, WAN, ...) in network byte order
		char		hostname_str[EPS_HOSTNAME_LENGTH];	///< first few characters of the hostname (should be long enough to cover most cases)
	};

	/// Communication header:  list all devices.
	struct CommList2 final
	{
		CommHeader1		header;				///< common header
		uint16_t		number_of_records;	///< number of records in the @ref data array (network byte order)
		CommList2Data	data[0];			///< array of devices

		/// Initialize the header.
		void set(const size_t records);
	};

	/// Communication header:  send stats information.
	struct CommStats1 final
	{
		CommHeader1		header;
		uint64_t		server_bytes_received;	///< total bytes received from the server
		uint64_t		server_bytes_sent;		///< total bytes sent to the server
		uint64_t		bytes_dropped;			///< total bytes dropped
		uint64_t		direct_bytes_received;	///< total direct client-to-client bytes received
		uint64_t		direct_bytes_sent;		///< total direct client-to-client bytes sent
		uint64_t		tap_bytes_received;		///< total bytes read from the TAP interface
		uint64_t		tap_bytes_sent;			///< total bytes written to the TAP interface

		uint32_t		timestamp;				///< when the stats packet was generated (network byte order)

		char			server_bps_in	[EPS::bps_string_size];	///< text representation of server rx bps
		char			server_bps_out	[EPS::bps_string_size];	///< text representation of server tx bps
		char			direct_bps_in	[EPS::bps_string_size];	///< text representation of direct (mesh) rx bps
		char			direct_bps_out	[EPS::bps_string_size];	///< text representation of direct (mesh) tx bps
		char			tap_bps_in		[EPS::bps_string_size];	///< text representation of tap interface rx bps
		char			tap_bps_out		[EPS::bps_string_size];	///< text representation of tap interface tx bps

		uint32_t		mesh_connections;		///< number of uuids in the array (mesh connections, network byte order)
		UUID16			uuid[0];				///< established mesh connections

		/// Initialize the record.
		void set(const UUIDSet &uuids);
	};

	#pragma pack(pop)

	/// Convert EPS::ECommType to a user-readable text string. @{
	std::string comm_type_to_string( const uint8_t type );
	std::string comm_type_to_string( const EPS::ECommType type );
	/// @}
}
